# 2012-10-12
# The aim is to read a set of sampled MUL-trees, find the 
# ways in which the tetraploids have been partitioned, and make new
# subsets of MUL-trees samples, one subset for each
# commonly occurring partition of tetraploids. Rare partitions
# can be just listed.



library(ape)
library(stringr)


fpathbase <- "C:/Users/Graham/AAA/Programming/Goteborg/AlloppRcode-nhybs/sampledmultrees"


glue <- function(...) {
  paste(...,sep="")
}


insist <- function(e, msg) {
  if (!e) {
    stop(msg)
  }
}
  


nexus.read.lines <- function(filepath) {
  tlines <- scan(filepath, what="", sep="\n", quiet=TRUE)
  tlines
}


nexus.read.header <- function(textlines) {
  insist(str_trim(textlines[1]) == "#NEXUS", "No '#NEXUS' found in file")
}


nexus.get.treeblock <- function(textlines) {
  begin.treeblock.index <- grep("BEGIN TREES;", x=textlines, ignore.case=TRUE)
  insist(length(begin.treeblock.index) == 1, "Found zero or more than one 'BEGIN TREES;' in file")
  treeblock <- textlines[(begin.treeblock.index +1):length(textlines)]
  end.treeblock.index <-  grep("END;|ENDBLOCK;", x=treeblock, ignore.case=TRUE)
  insist(length(end.treeblock.index) == 1, "Found zero or more than one 'END;' or 'ENDBLOCK;' in file")
  treeblock <- treeblock[1:(end.treeblock.index-1)]
  for (n in 1:length(treeblock)) {
    treeblock[n] <- str_trim(treeblock[n])
    }
  nonemptyline.indices <- which(nzchar(treeblock))
  treeblock <- treeblock[nonemptyline.indices]
  treeblock
}





nexus.get.translate.and.trees <- function(treeblock) {
  tree.indices <- grep("^TREE", x=treeblock, ignore.case=TRUE)
  translateblock <- treeblock[1:(tree.indices[1]-1)] 
  ntips <- length(translateblock) - 2
  stopifnot(grepl("TRANSLATE", translateblock[1], ignore.case=TRUE))
  stopifnot(grepl(";", translateblock[ntips+2], ignore.case=TRUE))
  translateblock <- translateblock[2:(ntips+1)]
  tipnameslist  <-  strsplit(translateblock, "[ \t,]")
  #print(tipnameslist)
  tipnames <- rep("", ntips)
  for (i in 1:ntips) {
    tipnames[i] <- tipnameslist[[i]][length(tipnameslist[[i]])]
    }
  treestrings <- treeblock[tree.indices]  
  list(tipnames=tipnames, treestrings=treestrings)
}


nexus.newicks.fromtreestrings <- function(treestrings) {
  openparens <- str_locate_all(treestrings, fixed("("))
  closeparens <- str_locate_all(treestrings, fixed(")"))
  newickstrings <- rep("", length(treestrings))
  for (tr in 1:length(treestrings)) {
    newickstrings[tr] <- substr(treestrings[tr], openparens[[tr]][1,"start"], closeparens[[tr]][nrow(closeparens[[tr]]),"end"])
    }
  newickstrings
}




# recursively builds a list of nodes from part of a Newick character vector
subcharvect.to.subtree <- function(charvect, subtree) {
paren <- 0
brack <- 0
comma <- -1
closep <- -1
openp <- -1
for (i in 1:length(charvect)) {
  if (charvect[i] == "(") { paren <- paren + 1 }
  if (charvect[i] == ")") { paren <- paren - 1 }
  if (charvect[i] == "[") { brack <- brack + 1 }
  if (charvect[i] == "]") { brack <- brack - 1 }
  stopifnot(brack <= 1)
  if (brack==0  &&  paren==1  &&  charvect[i] == ",") {
    comma <- i
    }
  if (charvect[i] == "(" &&  paren==1) {
    openp <- i
    }
  if (charvect[i] == ")" &&  paren==0) {
    closep <- i
    }
  }
stopifnot((comma > 0  &&  closep > 0  &&  openp > 0)  ||  (comma < 0  &&  closep < 0  &&  openp < 0))

opbs <- which(charvect=="[")
lastopb <- opbs[length(opbs)]
clbs <- which(charvect=="]")
lastclb <- clbs[length(clbs)]
colons <- which(charvect==":")
stopifnot(lastopb < lastclb)
brlen <- -1
if (length(colons) > 0) {
  lastcolon <- colons[length(colons)]
  if (lastcolon == lastclb+1) {
    brlen <- as.double(paste(charvect[(lastcolon+1):length(charvect)], sep="", collapse=""))
    } 
  }
annot <- paste(charvect[(lastopb+1):(lastclb-1)], sep="", collapse="")

if (comma < 0) {  
  subtree <- c((subtree), list(list(node=charvect[1:(lastopb-2)], brlen=brlen, annot=annot, lft=-1, rgt=-1)))
  } else {
  subtree <- subcharvect.to.subtree(charvect[2:(comma-1)], subtree)
  lfti <- length(subtree)     
  subtree <- subcharvect.to.subtree(charvect[(comma+1):(closep-1)], subtree)
  rgti <- length(subtree)  
  subtree <- c((subtree), list(list(node="", brlen=brlen, annot=annot, lft=lfti, rgt=rgti)))
  }
subtree
}               


# builds a list of nodes from a Newick character vector
charvect.to.nodelist <- function(tree.charvect) {
op <- which(tree.charvect=="(")
sc <- which(tree.charvect==";")
stopifnot(length(sc)==1)
charvect <- tree.charvect[op[1]:(sc-1)]
x <- subcharvect.to.subtree(charvect, NULL)
x
}



get.diploid.indices <- function(tipnames) {
  lastchars <- substr(tipnames, nchar(tipnames), nchar(tipnames))
  which(lastchars!="0"  &  lastchars!="1")
}

get.tetraploid.seq.indices <- function(tipnames) {
  lastchars <- substr(tipnames, nchar(tipnames), nchar(tipnames))
  which(lastchars=="0"  |  lastchars=="1")
}


make.tetraploid.species.map <- function(tipnames, tetseqindices) {
  tetnames <- substr(tipnames[tetseqindices], 1, nchar(tipnames[tetseqindices])-1)
  tetnames <- unique(tetnames)
  map <- rep(0, length(tipnames))
  for (n in 1:length(tipnames)) {
    tetname <- substr(tipnames[n], 1, nchar(tipnames[n])-1)
    for (i in 1:length(tetnames)) {
      if (tetname == tetnames[i]) {
        map[n] <- i
      }
    }
  }
  map
}


tetree.index.from.annot <- function(annot) {
  ttival <- -1
  tti <- str_locate(annot, "&tti=T")
  ttiend <- tti[1,"end"]
  if (is.finite(ttiend)) {
    ttivalstr <- substr(annot, ttiend+1, nchar(annot))
    comma <- str_locate(ttivalstr, ",")
    if (is.finite(comma[1,"start"])) {
      ttivalstr <- substr(ttivalstr, 1, comma-1)
    }
  ttival <- as.integer(ttivalstr) + 1
  }
ttival
}


tetraploids.of.node <- function(nodelist, n, tetsppmap, tetset) { 
  if (nodelist[[n]]$lft <= 0) {
    tetseqidx <- as.integer(nodelist[[n]]$node)
    tetset <- c(tetset, tetsppmap[tetseqidx])
  } else {
    lftset <- tetraploids.of.node(nodelist, nodelist[[n]]$lft, tetsppmap, integer(0))
    rgtset <- tetraploids.of.node(nodelist, nodelist[[n]]$rgt, tetsppmap, integer(0))
    tetset <- c(tetset, lftset, rgtset)
  }
tetset
}


get.matrixof.tetsetsets <- function(newickstrings, noftets) {
  matrix.tetsetsets <- matrix("", nrow=length(newickstrings), ncol=noftets)
  for (tr in 1:length(newickstrings)) {
    cat(".")
    if (tr %% 100 == 0) { cat("\n") }
    newickascharvect <- unlist(strsplit(newickstrings[tr], "")) 
    nodelist <- subcharvect.to.subtree(newickascharvect, NULL)
    doneflags <- rep(FALSE, 2*noftets)
    tetsetsets <- character(0)
    for (n in 1:length(nodelist)) {
      annot <- nodelist[[n]]$annot
      tti <- tetree.index.from.annot(annot)
      #cat("n=", n, "tti=", tti, doneflags, "\n")
      if (tti > 0  &&  !doneflags[tti]) {
        doneflags[tti] <- TRUE
        tetsetidxs <- tetraploids.of.node(nodelist, n, tetsppmap, integer(0))
        tetsetidxs <- sort(tetsetidxs)
        tetsetstring <- paste(as.character(tetsetidxs), collapse="-")
        tetsetsets <- c(tetsetsets, tetsetstring)
      }
    }
    matrix.tetsetsets[tr,1:length(tetsetsets)] <- sort(tetsetsets)
  }
  cat("\n")
  matrix.tetsetsets 
}




tetsetsets.equal <- function(a, b) {
  if (length(a) != length(b)) {
    return (FALSE)
  }
  
}

tetsetsets.occurring <- function(list.tetsetsets) {
tsso <- list(list.tetsetsets[[1]])
for (i in 2:length(list.tetsetsets)) {
  tetsetsets <- list.tetsetsets[[i]]
  for (j in 1:length(tsso)) {
    tso <- tsso[[j]]
    }
  }
}



write.nexus <- function(textlines, treelines, treestrings, matrix.tetsetsets, tetsetsets) {
  tr <- 1
  for (n in 1:length(textlines)) {
    if (treelines[n]) {
      if (identical(matrix.tetsetsets[tr,], tetsetsets))
      cat(treestrings[tr], "\n")
      tr <- tr+1
    } else {
      cat(textlines[n], "\n")
    }
  }
}


sink.nexus <- function(fpath, textlines, treelines, treestrings, matrix.tetsetsets, tetsetsets) {
sink(file=fpath)
write.nexus(textlines, treelines, treestrings, matrix.tetsetsets, tetsetsets)
sink(NULL)
}



fpath <- glue(fpathbase, ".txt")

# read in nexus file, check it is a nexus file
textlines <- nexus.read.lines(fpath)
nexus.read.header(textlines)

# read the tree block, example with dips Salsv, Suw; tets Si, Ss, St:

#  Begin trees;
#  	Translate
#  		1 Salsv,
#  		2 Si0,
#  		3 Si1,
#  		4 Ss0,
#  		5 Ss1,
#  		6 St0,
#  		7 St1,
#  		8 Suw
#  		;
#  tree STATE_0 = [&R] ((((4:[&tti=X]0.114,2:[&tti=X]0.114):[&tti=T0]0.010,6:[&tti=T1]0.125):[&tti=X]0.058,(((5:[&tti=X]0.114,3:[&tti=X]0.114):[&tti=T0]0.006,7:[&tti=T1]0.121):[&tti=X]0.008,1:[&tti=X]0.130):[&tti=X]0.053):[&tti=X]0.014,8:[&tti=X]0.198);
#  tree STATE_1000 = [&R] (8:[&tti=X]0.005,((7:[&tti=T1]0.005,((4:[&tti=T0]0.003,(1:[&tti=X]4.7E-4,5:[&tti=T0]4.7E-4):[&tti=X]0.003131246478211614):[&tti=X]0.001,(3:[&tti=T2]0.004,2:[&tti=T2]0.004):[&tti=X]1.3E-5):[&tti=X]2.9E-4):[&tti=X]3.1E-6,6:[&tti=T1]0.005):[&tti=X]1.6E-4);
#  ...
#  tree STATE_99000 = [&R] (((((4:[&tti=X]4.5E-4,2:[&tti=X]4.5E-4):[&tti=T1]6.2E-4,6:[&tti=T0]0.001):[&tti=X]6.9E-6,1:[&tti=X]0.001):[&tti=X]4.2E-6,(5:[&tti=X]4.5E-4,3:[&tti=X]4.5E-4):[&tti=T1]6.3E-4):[&tti=X]9.6E-4,(8:[&tti=X]0.001,7:[&tti=T0]0.001):[&tti=X]4.0E-4);
#  End;
treelines <- grepl("^TREE", x=textlines, ignore.case=TRUE)
treeblock <- nexus.get.treeblock(textlines)
transandtrees <- nexus.get.translate.and.trees(treeblock)

# tipnames is a vector of tip names: eg "Salsv" "Si0"   "Si1"   "Ss0"   "Ss1"   "St0"   "St1"   "Suw" 
tipnames <- transandtrees$tipnames

# treestrings is a vector of text lines conating one Newick string each.
treestrings <- transandtrees$treestrings

# dipindices is the tip indices in the nexus file of the 
# diploid tips, eg 1 8
dipindices <- get.diploid.indices(tipnames)

# tetseqindices is the tip indices in the nexus file of the 
# tetraploid tips (two per tetraploid species). eg  2 3 4 5 6 7
tetseqindices <- get.tetraploid.seq.indices(tipnames)    # tetseqindices
noftets <- length(tetseqindices)/2

# tetsppmap maps tip indices to tetraploid species indices
# tetsppmap[i] == 0 means tip index i is not a tetraploid.
# tetsppmap[i] == j means tip index i is the j'th tetraploid species.
# Eg 0 1 1 2 2 3 3 0. No guarantee about ordering of j
tetsppmap <- make.tetraploid.species.map(tipnames, tetseqindices)

# Extract Newick strings, minus final ;
newickstrings <- nexus.newicks.fromtreestrings(treestrings)

# Gets a list of the partitions of tetraploids into subtrees.
# This can be slow.
matrix.tetsetsets <- get.matrixof.tetsetsets(newickstrings, noftets)
# Example of result:
#    [1,] "1-2" "3"  ""  
#    [2,] "1"   "2"  "3" 
#    ....
#  [100,] "1-2" "3"  "" 

unique.tetsetsets <- unique(matrix.tetsetsets)
for (i in 1:nrow(unique.tetsetsets)) {
  w <- unique.tetsetsets[i,]
  w <- w[nchar(w)>0]
  filesuffix <- paste(w, collapse="_")
  sink.nexus(glue(fpathbase, filesuffix, ".txt"), textlines, treelines, treestrings, matrix.tetsetsets, unique.tetsetsets[i,])
  }