# 2012-09-25
# Aimed at reading annotated trees as produced by TreeAnnotator,
# then summarizing values stored as a discrete sample, 
# and re-writing in a way that FigTree can display. Some nodes in the
# concensus tree will have some hybridization heights, but there
# will be missing values. I'd like eg a median of those that are there.
# I will encode all values as strings.

# It will fail to read some nexus files containing tree(s).
# It is inefficient.


library(ape)
library(stringr)




# extracts a tree from a nexus file as a character vector 
# (ie text string split into array)
tree.from.nexus <- function(filepath) {
x <- scan(filepath, what="", sep="\n", quiet=TRUE)
begin.trees.index <- grep("BEGIN TREES;", x=x, ignore.case=TRUE)
all.trees <- x[(begin.trees.index +1):length(x)]
end.trees.index <-  grep("END;|ENDBLOCK;", x=all.trees, ignore.case=TRUE)
all.trees <- all.trees[1:(end.trees.index-1)]
first.tree.index <- grep("TREE", x=all.trees, ignore.case=TRUE)
first.tree <- all.trees[first.tree.index]
translate.block <- all.trees[1:(first.tree.index-1)] 
#print(translate.block)
ntips <- length(translate.block) - 2
stopifnot(grepl("TRANSLATE", translate.block[1], ignore.case=TRUE))
stopifnot(grepl(";", translate.block[ntips+2], ignore.case=TRUE))
translate.block <- translate.block[2:(ntips+1)]
tipnameslist  <-  strsplit(translate.block, "[ \t,]")
#print(tipnameslist)
tipnames <- rep("", ntips)
for (i in 1:ntips) {
  tipnames[i] <- tipnameslist[[i]][length(tipnameslist[[i]])]
  }
tree.charvect <- unlist(strsplit(first.tree, ""))
list(tree.string=first.tree, tree.charvect=tree.charvect, tipnames=tipnames)
}



# recursively builds a list of nodes from part of a Newick character vector
subcharvect.to.subtree <- function(charvect, subtree) {
paren <- 0
brack <- 0
comma <- -1
closep <- -1
openp <- -1
for (i in 1:length(charvect)) {
  if (charvect[i] == "(") { paren <- paren + 1 }
  if (charvect[i] == ")") { paren <- paren - 1 }
  if (charvect[i] == "[") { brack <- brack + 1 }
  if (charvect[i] == "]") { brack <- brack - 1 }
  stopifnot(brack <= 1)
  if (brack==0  &&  paren==1  &&  charvect[i] == ",") {
    comma <- i
    }
  if (charvect[i] == "(" &&  paren==1) {
    openp <- i
    }
  if (charvect[i] == ")" &&  paren==0) {
    closep <- i
    }
  }
stopifnot((comma > 0  &&  closep > 0  &&  openp > 0)  ||  (comma < 0  &&  closep < 0  &&  openp < 0))

opbs <- which(charvect=="[")
lastopb <- opbs[length(opbs)]
clbs <- which(charvect=="]")
lastclb <- clbs[length(clbs)]
colons <- which(charvect==":")
stopifnot(lastopb < lastclb)
brlen <- -1
if (length(colons) > 0) {
  lastcolon <- colons[length(colons)]
  if (lastcolon == lastclb+1) {
    brlen <- as.double(paste(charvect[(lastcolon+1):length(charvect)], sep="", collapse=""))
    } 
  }
annot <- paste(charvect[(lastopb+1):(lastclb-1)], sep="", collapse="")

if (comma < 0) {  
  subtree <- c((subtree), list(list(node=charvect[1:(lastopb-1)], brlen=brlen, annot=annot, lft=-1, rgt=-1)))
  } else {
  subtree <- subcharvect.to.subtree(charvect[2:(comma-1)], subtree)
  lfti <- length(subtree)     
  subtree <- subcharvect.to.subtree(charvect[(comma+1):(closep-1)], subtree)
  rgti <- length(subtree)  
  subtree <- c((subtree), list(list(node="", brlen=brlen, annot=annot, lft=lfti, rgt=rgti)))
  }
subtree
}               


# builds a list of nodes from a Newick character vector
charvect.to.nodelist <- function(tree.charvect) {
op <- which(tree.charvect=="(")
sc <- which(tree.charvect==";")
stopifnot(length(sc)==1)
charvect <- tree.charvect[op[1]:(sc-1)]
x <- subcharvect.to.subtree(charvect, NULL)
x
}


 
# nodelist to nodematrix: loses annotations
nodelist.to.nodematrix <- function(nodelist, tipnames) {
nofnodes <- length(nodelist)
nodenames  <- rep("", nofnodes)
for (i in 1:nofnodes) {
  if (nchar(nodelist[[i]]$node) > 0) {
    ni <- tipnames[ as.integer(nodelist[[i]]$node) ]
    } else {
    ni <- "*"
    }
  nodenames[i] <- ni
  }
nodes <- matrix(-1, nrow=nofnodes, ncol=5, dimnames=list(nodenames,c("node", "lft", "rgt", "brlen", "posterior")))

for (i in 1:nofnodes) {
  nodes[i,1] <- i
  nodes[i,2] <- as.integer(nodelist[[i]]$lft)
  nodes[i,3] <- as.integer(nodelist[[i]]$rgt)
  nodes[i,4] <- as.double(nodelist[[i]]$brlen)
  vals <- unlist(strsplit(nodelist[[i]]$annot, ","))
  p.i <- grep("posterior", vals)
  if (length(p.i) > 0) {
    #print(vals[p.i])
    p.s <- substr(vals[p.i], 11, nchar(vals[p.i]))
    p.v <- as.double(p.s)
    } else {
    p.v <- -1
    }
  nodes[i,5] <- p.v 
  }
nodes
}


# nodematrix to newick: no annotations. OK for ape
nodematrix.to.newick <- function(nodematrix, n, txt) {
if (nodematrix[n,"lft"] < 0) {
  stopifnot(nodematrix[n,"rgt"] < 0)
  txt <- paste(txt, rownames(nodematrix)[n], ":", nodematrix[n,"brlen"], sep="")
  } else {
  lft <- nodematrix.to.newick(nodematrix, nodematrix[n,"lft"], txt)
  rgt <- nodematrix.to.newick(nodematrix, nodematrix[n,"rgt"], txt)
  brlen <- round(nodematrix[n,"brlen"], digits=8)
  if (brlen >= 0) {   # root has brlen=-1
    txt <- paste("(", lft, ",", rgt, "):", brlen, sep="")
    } else {
    txt <- paste("(", lft, ",", rgt, ");", sep="")
    }
  }
txt
}


# nodelist to newick. Contains annotations
nodelist.to.newick <- function(nodelist, n, txt) {
if (nodelist[[n]]$lft < 0) {
  stopifnot(nodelist[[n]]$rgt < 0)
  txt <- paste(txt, nodelist[[n]]$node, "[", nodelist[[n]]$annot, "]:", nodelist[[n]]$brlen, sep="")
  } else {
  lft <- nodelist.to.newick(nodelist, nodelist[[n]]$lft, txt)
  rgt <- nodelist.to.newick(nodelist, nodelist[[n]]$rgt, txt)  
  brlen <- nodelist[[n]]$brlen
  if (brlen >= 0) {   # root has brlen=-1
    txt <- paste("(", lft, ",", rgt, ")[", nodelist[[n]]$annot, "]:", brlen, sep="")
    } else {
    txt <- paste("(", lft, ",", rgt, ")", nodelist[[n]]$annot, ";", sep="")
    }
  }
txt
}


fpath <- "C:/Users/Graham/AAA/Programming/GoteborgSim-fromdesktop/2012-09-04/G/Gg3n1t80/G5MTCg3n1t80r3.txt"

                                     
tree.tips.txt <- tree.from.nexus(fpath)
nodelist <- charvect.to.nodelist(tree.tips.txt$tree.charvect)
annotnewick <- nodelist.to.newick(nodelist, length(nodelist), "")
sink(file = paste(fpath, "R.txt", sep=""))
cat(annotnewick)
sink(NULL)
nodematrix <- nodelist.to.nodematrix(nodelist, tree.tips.txt$tipnames)
#newick <- nodematrix.to.newick(nodematrix, nrow(nodematrix), "")
#phylo <- read.tree(text=newick)
#plot.phylo(phylo)



####################################
# gets substrings like 
# {"T1","X","T0"}
# {0.0378,0.9404,0.0218}
# from annotations

extract_set <- function(setname, string) {
  stopifnot(length(string)==1)
  set <- str_locate_all(paste(setname, "=[{]", sep=""), string=string)[[1]]
  close.curly <- str_locate_all("[}]", string=string)[[1]]
  endi <- -1
  nr <- nrow(close.curly)
  stopifnot(nr > 0)
  for (i in 1:nr) {
    if (endi < 0  &&  set[1,"end"] < close.curly[i,"start"])
      {
      endi <- i
      }
    }
  stopifnot(endi > 0)
  substr(string, set[1,"end"], close.curly[endi,"start"])
}


tti.set <- extract_set("tti.set", nodelist[[6]]$annot)
tti.set.prob <- extract_set("tti.set.prob", nodelist[[6]]$annot)


cat(tti.set)
cat(tti.set.prob)


# do something like above for hybrid heights, but the set will be
# lots of numbers/NaNs as strings, eg "h0.0138", "hNA"